#!/usr/local/bin/python

# Copyright (c) 2007 The Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions, and the following disclaimer.
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions, and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#   3. Redistributions must acknowledge that this software was
#      originally developed by the UCSF Computer Graphics Laboratory
#      under support by the NIH National Center for Research Resources,
#      grant P41-RR01081.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time
import getpass
import math

import parser

SymbolDict = {
	"0":		None,
	"none":		None,
	"1":		"circle",
	"circle":	"circle",
	"2":		"square",
	"square":	"square",
	"3":		"hexagon",
	"hexagon":	"hexagon",
	"4":		"uparrow",
	"uparrow":	"uparrow",
	"5":		"downarrow",
	"downarrow":	"downarrow",
}

def readTopo(inFile):
	if type(inFile) is type(''):
		inFile = open(inFile)
	p = parser.TopoParser()
	p.feed(inFile.read())
	p.close()
	return p.topo

def topops(topo, outFile, frame, big):
	if type(outFile) is type(''):
		f = open(outFile, "w")
	else:
		f = outFile

	#
	# Prolog
	#
	f.write("%!PS-Adobe-2.0\n")
	f.write("%%%%Title: %s transmembrane topology\n" % topo.name)
	f.write("%%Creator: topops\n")
	f.write("%%%%CreationDate: %s\n" % time.ctime(time.time()))
	f.write("%%%%For: %s\n" % getpass.getuser())
	f.write("%%DocumentFonts: Helvetica\n")
	f.write("%%Pages: 1 1\n")
	f.write("%%EndComments\n")
	writeProcSet(f)
	f.write("%%EndProlog\n")
	f.write("%%BeginSetup\n")
	f.write("setupFonts\n")
	f.write("%%EndSetup\n")
	f.write("%%Page: 1 1\n")
	f.write("%%PageFonts: Helvetica\n")
	xmin, xmax, ymin, ymax = bbox(topo)
	xSize = float(xmax - xmin)
	ySize = float(ymax - ymin)
	xuScale = (6.5 * 72) / xSize	# unrotated scaling
	yuScale = (9.0 * 72) / ySize
	uScale = min(xuScale, yuScale)
	xrScale = (6.5 * 72) / ySize	# rotated scaling
	yrScale = (9.0 * 72) / xSize
	rScale = min(xrScale, yrScale)
	if rScale > uScale:
		bb = (540 - ySize * rScale, 72, 540, xSize * rScale + 72)
	else:
		bb = (72, 72, xSize * uScale + 72, ySize * uScale + 72)
	f.write("%%%%PageBoundingBox: %g %g %g %g\n" % bb)
	f.write("%%BeginPageSetup: 1 1\n")
	f.write("%%PaperSize: Letter\n")
	if frame:
		f.write("0 setgray\n")
		f.write("%g %g moveto\n" % (bb[0], bb[1]))
		f.write("%g %g lineto\n" % (bb[2], bb[1]))
		f.write("%g %g lineto\n" % (bb[2], bb[3]))
		f.write("%g %g lineto\n" % (bb[0], bb[3]))
		f.write("closepath stroke\n")
	if rScale > uScale:
		f.write("540 72 translate\n")
		f.write("90 rotate\n")
		f.write("%g %g scale\n" % (rScale, rScale))
		f.write("%g %g translate\n" % (-xmin, -ymin))
	else:
		f.write("72 72 translate\n")
		f.write("%g %g scale\n" % (uScale, uScale))
		f.write("%g %g translate\n" % (-xmin, -ymin))
	f.write("%%EndPageSetup: 1 1\n")
	ps(f, topo, xmin, xmax, ymin, ymax, big)
	f.write("showpage\n")
	f.write("%%PageTrailer\n")
	f.write("%%Trailer\n")

def topoeps(topo, outFile, frame):
	if type(outFile) is type(''):
		f = open(outFile, "w")
	else:
		f = outFile

	#
	# Prolog
	#
	f.write("%!PS-Adobe-2.0 EPSF-2.0\n")
	f.write("%%%%Title: %s transmembrane topology\n" % topo.name)
	f.write("%%Creator: topoeps\n")
	f.write("%%%%CreationDate: %s\n" % time.ctime(time.time()))
	f.write("%%%%For: %s\n" % getpass.getuser())
	f.write("%%DocumentFonts: Helvetica\n")
	xmin, xmax, ymin, ymax = bbox(topo)
	f.write("%%%%BoundingBox: %d %d %d %d\n"
			% (xmin, ymin, xmax, ymax))
	f.write("%%EndComments\n")
	writeProcSet(f)
	f.write("%%EndProlog\n")
	f.write("%%BeginSetup\n")
	f.write("setupFonts\n")
	if frame:
		f.write("0 setgray\n")
		f.write("%g %g moveto\n" % (xmin, ymin))
		f.write("%g %g lineto\n" % (xmin, ymax))
		f.write("%g %g lineto\n" % (xmax, ymax))
		f.write("%g %g lineto\n" % (xmax, ymin))
		f.write("closepath stroke\n")
	f.write("%%EndSetup\n")
	ps(f, topo, xmin, xmax, ymin, ymax, big)
	f.write("%%Trailer\n")

def bbox(topo):
	xmin = int(math.floor(topo.xMin))
	ymin = int(math.floor(topo.yMin))
	xmax = int(math.ceil(topo.xMax))
	ymax = int(math.ceil(topo.yMax))
	if topo.labelTop:
		ymax = ymax + 12
	if topo.labelBottom:
		ymin = ymin - 12
	return xmin, xmax, ymin, ymax

def ps(f, topo, xmin, xmax, ymin, ymax, big):
	#
	# Top and bottom labels
	#
	f.write("0 setgray\n")
	if topo.labelTop:
		f.write("useLargeFont\n")
		f.write("%g %g %g (%s) centerTop\n"
			% (xmin, xmax, ymax, topo.labelTop))
	if topo.labelBottom:
		f.write("useLargeFont\n")
		f.write("%g %g %g (%s) centerBottom\n"
			% (xmin, xmax, ymin, topo.labelBottom))

	#
	# Membrane borders
	#
	f.write("%g %g moveto %g %g lineto\n" %
		(topo.xMembMin, topo.yMembTop, topo.xMembMax, topo.yMembTop))
	f.write("%g %g moveto %g %g lineto\n" %
		(topo.xMembMin, topo.yMembBottom,
		topo.xMembMax, topo.yMembBottom))
	f.write("stroke\n")

	#
	# Amino acids
	#
	f.write("useSmallFont\n")
	topoagain = []
	for loc in topo.locations:
		good = 1
		cmd = SymbolDict[loc.symbol]
		if loc.border != loc.foreground:
			topoagain.append(loc)
			good = 0
		if cmd and good == 1:
			f.write("newpath %g %g %g %s "
				% (loc.x, loc.y, topo.radius, cmd))
			bg = parser.str2rgb(loc.background)
			if not isWhite(bg):
				f.write("gsave %s fill grestore " % rgb(bg))
			f.write("%s stroke\n" %
				rgb(parser.str2rgb(loc.border)))
		f.write("%s %g %g (%s) center\n" %
				(rgb(parser.str2rgb(loc.foreground)),
				loc.x, loc.y, loc.type))
	for loc in topoagain:
		cmd = SymbolDict[loc.symbol]
		if big == 1:
			topo.radius = 7.5
		if cmd:
			f.write("newpath %g %g %g %s "
				% (loc.x, loc.y, topo.radius, cmd))
			bg = parser.str2rgb(loc.background)
			if not isWhite(bg):
				f.write("gsave %s fill grestore " % rgb(bg))
			f.write("%s stroke\n" %
				rgb(parser.str2rgb(loc.border)))
		f.write("%s %g %g (%s) center\n" %
				(rgb(parser.str2rgb(loc.foreground)),
				loc.x, loc.y, loc.type))

	# Zero bridges
	#
	try:
		zbList = topo.zerobridge
	except AttributeError:
		pass
	else:
		for zb in zbList:
			fg = [0, 0, 0]
			f.write("%s newpath " % rgb(fg))
			cmd = "moveto"
			for pt in zb:
				f.write("%g %g %s " % (pt[0], pt[1], cmd))
				cmd = "lineto"
			f.write("stroke\n")

	# feature boxes
	#
	try:
		fbList = topo.feature
	except AttributeError:
		pass
	else:
		for fb in fbList:
			fg = [0, 0, 0]
			f.write("%s newpath " % rgb(fg))
			cmd = "moveto"
			for pt in fb:
				f.write("%g %g %s " % (pt[0], pt[1], cmd))
				cmd = "lineto"
			f.write("stroke\n")


def isWhite(rgb):
	return rgb[0] == 1 and rgb[1] == 1 and rgb[2] == 1

def rgb(n):
	b = n[2]
	g = n[1]
	r = n[0]
	if r == g and g == b:
		return "%g setgray" % r
	else:
		return "%g %.3f %g setrgbcolor" % (r, g, b)

def writeProcSet(f):
	f.write(ProcSet)

ProcSet = """
%%BeginProcSet: topo

% setupFonts: find and scale all fonts
/setupFonts {					% -
	/Helvetica findfont			% font
	dup 7 scalefont /smallFont exch def	% font
	14 scalefont /largeFont exch def	% -
} def

% useLargeFont: make large font default
/useLargeFont {
	largeFont setfont
} def

% useSmallFont: make small font default
/useSmallFont {
	smallFont setfont
} def

% stringSize: get offset and size (relative to origin) of a string bbox
/stringSize {					% s
	newpath 0 0 moveto			% s
	false charpath				% -
	flattenpath pathbbox			% llx lly urx ury
	newpath					% llx lly urx ury
} def

% centerTop: center a string given x limits and top y position
/centerTop {					% minx maxx maxy s
	/s exch def				% minx maxx maxy
	/maxy exch def				% minx maxx
	add 2 div				% cx
	s stringSize				% cx llx lly urx ury
	/ury exch def /urx exch def		% cx llx lly
	/lly exch def /llx exch def		% cx
	urx llx sub 2 div sub			% x
	maxy ury sub				% x y
	moveto s show				% -
} def

% centerBottom: center a string given x limits and bottom y position
/centerBottom {					% minx maxx miny s
	/s exch def				% minx maxx maxy
	/miny exch def				% minx maxx
	add 2 div				% cx
	s stringSize				% cx llx lly urx ury
	/ury exch def /urx exch def		% cx llx lly
	/lly exch def /llx exch def		% cx
	urx llx sub 2 div sub			% x
	miny lly sub				% x y
	moveto s show				% -
} def

% center: center a string at an (x, y) location
/center {					% cx cy s
	/s exch def				% cx cy
	/cy exch def				% cx
	/cx exch def				% -
	s stringSize				% llx lly urx ury
	/ury exch def /urx exch def		% llx lly
	/lly exch def /llx exch def		% -
	cx urx llx add 2 div sub		% x
	cy ury lly add 2 div sub		% x y
	moveto s show				% -
} def

% circle: draw circle of given radius at (x, y) location
/circle {					% x y r
	0 360 arc closepath
} def

% square: draw square at (x, y) location
/square {					% x y r
	/r exch def				% x y
	/d r 2 mul def				% x y
	/y exch def				% x
	/x exch def				% -
	x r sub y r sub moveto			% -
	0 d rlineto				% -
	d 0 rlineto				% -
	0 d neg rlineto				% -
	closepath				% -
} def

% hexagon: draw hexagon at (x, y) location
/hexagon {
	/r exch def				% x y
	/dr r 2 div def				% x y
	/y exch def				% x
	/x exch def				% -
	x y r add moveto			% -
	x r add y dr add lineto			% -
	x r add y dr sub lineto			% -
	x y r sub lineto			% -
	x r sub y dr sub lineto			% -
	x r sub y dr add lineto			% -
	closepath				% -
} def

% uparrow: draw up arrow at (x, y) location
/uparrow {
	/r exch def				% x y
	/dr r 2 div def				% x y
	/y exch def				% x
	/x exch def				% -
	x y r add moveto			% -
	x r add y dr sub lineto			% -
	x y r sub lineto			% -
	x r sub y dr sub lineto			% -
	closepath				% -
} def

% downarrow: draw down arrow at (x, y) location
/downarrow {
	/r exch def				% x y
	/dr r 2 div def				% x y
	/y exch def				% x
	/x exch def				% -
	x y r add moveto			% -
	x r add y dr add lineto			% -
	x y r sub lineto			% -
	x r sub y dr add lineto			% -
	closepath				% -
} def

%%EndProcSet: topo
"""

if __name__ == "__main__":
	def usage(prog):
		print "Usage: %s [-e] [-b] [-o file.ps] [file.xml]" % prog
		print "\t-e:\tgenerate Encapsulated PostScript"
		print "\t-o:\tsend output to file.ps"
		raise SystemExit, 1
	import sys
	import getopt
	frame = 0
	big = 0
	converter = topops
	output = sys.stdout
	try:
		opts, args = getopt.getopt(sys.argv[1:], "ebfo:")
	except:
		usage(sys.argv[0])
	for opt, val in opts:
		if opt == "-e":
			converter = topoeps
		elif opt == "-f":
			frame = 1
		elif opt == "-b":
			big = 1
		elif opt == "-o":
			output = val
	if len(args) == 0:
		input = sys.stdin
	elif len(args) == 1:
		input = args[0]
	else:
		usage(sys.argv[0])
	converter(readTopo(input), output, frame, big)

	#topoeps(readTopo("oct2.xml"), "oct2.eps", 0)
	#topops(readTopo("oct2.xml"), "oct2.ps", 1)
